#include <Configuration.h>
#include <BasicAbstractions/Assets.h>
#include <States/GameState/Entities/DynamicEntities/DynamicEntities.h>

Enemy::Enemy(const float speed) : m_speed(speed) {
    std::mt19937 gen{std::random_device{}()};
    std::uniform_int_distribution<size_t> dist{0, 3};
    const sf::Texture* texture;
    switch (dist(gen)) {
        case 0: texture = &Assets::GetEnemyRedTexture(); break;
        case 1: texture = &Assets::GetEnemyYellowTexture(); break;
        case 2: texture = &Assets::GetEnemyBlueTexture(); break;
        case 3: texture = &Assets::GetEnemyPinkTexture(); break;
        default: texture = &Assets::GetEnemyRedTexture(); break;;
    }

    m_sprite.setTexture(*texture);
    m_sprite.setTextureRect(sf::IntRect(0, 0,
        static_cast<int>(texture->getSize().x / config::ENEMY_ON_SPRITE_COUNT), static_cast<int>(texture->getSize().y)));
    m_sprite.setOrigin(static_cast<float>(texture->getSize().x) / (2.f * config::ENEMY_ON_SPRITE_COUNT), static_cast<float>(texture->getSize().y) / 2.f);
    m_distance_limit = m_speed / config::ROOM_SIZE;
}

std::unique_ptr<IDynamicEntity> Enemy::clone() const {
    return std::make_unique<Enemy>(*this);
}

void Enemy::action() {
    if (m_is_moving) {
        const sf::Vector2f cur_pos = m_sprite.getPosition();
        const sf::Vector2f displacement = m_estimated_position - cur_pos;
        const float distance = std::sqrt(displacement.x * displacement.x + displacement.y * displacement.y);
        if (distance < m_distance_limit) {
            m_sprite.setPosition(m_estimated_position);
            m_is_moving = false;
            return;
        }
        const float frame_duration = m_stopwatch.getElapsedTime().asSeconds();
        const sf::Vector2f direction = m_speed * frame_duration * displacement / distance;
        float delta_x, delta_y;
        if (direction.x > 0)
            delta_x = std::min(direction.x, displacement.x);
        else
            delta_x = std::max(direction.x, displacement.x);
        if (direction.y > 0)
            delta_y = std::min(direction.y, displacement.y);
        else
            delta_y = std::max(direction.y, displacement.y);
        m_sprite.move(delta_x, delta_y);
        m_time_count += frame_duration;
        if(m_time_count > 1) {
            sf::IntRect rect = m_sprite.getTextureRect();
            rect.left = static_cast<int>((rect.left / rect.width + 4) % config::ENEMY_ON_SPRITE_COUNT) * rect.width;
            m_sprite.setTextureRect(rect);
            m_time_count = 0;
        }
    }
    else {
        if (static_cast<size_t>(m_stopwatch.getElapsedTime().asMilliseconds()) < m_dist_milliseconds(m_rng))
            return;
        const auto direction = static_cast<Room::Direction>(m_dist_direction(m_rng));
        m_ptr_room->get_side(direction)->enter(this);
    }
    m_stopwatch.restart();
}

void Enemy::draw_into(sf::RenderWindow& window) const {
    window.draw(m_sprite);
}

void Enemy::prepare_for_first_drawing() {
    m_sprite.setPosition(m_ptr_room->get_position());
}

void Enemy::prepare_for_drawing() {
    m_estimated_position = m_ptr_room->get_position();
    m_is_moving = true;
}